use crate::config::{Context, WebSocketCfg};
use crate::ip::IpEvent;
use crate::utils::{gethostname, sleep_s};

use serde_json::json;
use std::io::{self, Write};
use std::net::TcpStream;
use std::sync::Arc;
use tokio::sync::mpsc::Sender;
use tokio::sync::RwLock;
use tungstenite::stream::*;
use tungstenite::*;

pub async fn websocketreqrep(
    ctxarc: &Arc<RwLock<Context>>,
) -> WebSocket<MaybeTlsStream<TcpStream>> {
    let (mut wssocketrr, bootstrap_event);
    {
        let ctx = ctxarc.read().await;
        bootstrap_event = ctxarc.read().await.cfg.bootstrap_event().clone();

        wssocketrr = websocketconnect(&ctx.cfg.ws.get("reqrep").unwrap(), ctx.hostname.clone())
            .await
            .unwrap();
        send_to_ipbl_websocket(&mut wssocketrr, &bootstrap_event).await;
    }

    return wssocketrr;
}

pub async fn websocketpubsub(
    ctxarc: &Arc<RwLock<Context>>,
    txpubsub: Arc<RwLock<Sender<IpEvent>>>,
) {
    let ctx = ctxarc.read().await;
    let cfg = ctx.cfg.ws.get("pubsub").unwrap().clone();
    let hostname = ctx.hostname.clone();
    drop(ctx);
    let mut websocket = Arc::new(RwLock::new(
        websocketconnect(&cfg, hostname.clone()).await.unwrap(),
    ));
    tokio::spawn(async move {
        loop {
            let mut ws = websocket.write().await;
            match ws.read_message() {
                Ok(msg) => {
                    let tosend: IpEvent = match serde_json::from_str(msg.to_string().as_str()) {
                        Ok(o) => o,
                        Err(_e) => {
                            continue;
                        }
                    };
                    if tosend.ipdata.hostname != gethostname(true)
                        || tosend.msgtype == "init".to_string()
                    {
                        let txps = txpubsub.write().await;
                        txps.send(tosend).await.unwrap();
                    }
                }
                Err(e) => {
                    println!("error in pubsub: {e:?}");
                    ws.close(None).unwrap();
                    drop(ws);
                    websocket = Arc::new(RwLock::new(
                        websocketconnect(&cfg, hostname.clone()).await.unwrap(),
                    ));
                }
            };
        }
    });
}

pub async fn websocketconnect<'a>(
    wscfg: &WebSocketCfg,
    hostname: String,
) -> Result<WebSocket<MaybeTlsStream<TcpStream>>, Error> {
    print!("connecting to {} ... ", &wscfg.endpoint);
    io::stdout().flush().unwrap();
    let mut socket;
    loop {
        (socket, _) = match connect(&wscfg.endpoint) {
            Ok((o, e)) => (o, e),
            _ => {
                println!("error connecting to {}, retrying", &wscfg.endpoint);
                sleep_s(1).await;
                continue;
            }
        };
        break;
    }
    println!("connected to {}", &wscfg.endpoint);
    let msg = json!({ "hostname": hostname });
    socket
        .write_message(Message::Text(msg.to_string()))
        .unwrap();
    Ok(socket)
}

pub async fn send_to_ipbl_websocket(
    ws: &mut WebSocket<MaybeTlsStream<TcpStream>>,
    ip: &IpEvent,
) -> bool {
    let msg = format!("{val}", val = serde_json::to_string(&ip).unwrap());

    if ws.can_write() {
        match ws.write_message(Message::Text(msg)) {
            Ok(_) => {}
            Err(e) => {
                println!("err send read: {e:?}");
                return handle_websocket_error(ws);
            }
        };
    } else {
        return handle_websocket_error(ws);
    };

    if ws.can_read() {
        match ws.read_message() {
            Ok(_) => {}
            Err(e) => {
                println!("err send read: {e:?}");
                return handle_websocket_error(ws);
            }
        };
    } else {
        return handle_websocket_error(ws);
    };
    true
}

fn handle_websocket_error(ws: &mut WebSocket<MaybeTlsStream<TcpStream>>) -> bool {
    ws.close(None).unwrap();
    return false;
}
